@@ -20,11 +20,11 @@ import copy
 import operator
 import trafaret as t
 
-from bson import DBRef
-
 from flask import abort
 from flask.signals import _signals
 
+from importlib import import_module
+
 from pymongo import Connection, ASCENDING
 from pymongo.cursor import Cursor
 from pymongo.database import Database
@@ -36,9 +36,6 @@ from pymongo.son_manipulator import (SONManipulator, AutoReference,
 # list of collections for models witch need autoincrement id
 inc_collections = set([])
 
-# list of collections for models witch need auto dbref
-autoref_collections = {}
-
 after_insert = 'after_insert'
 after_update = 'after_update'
 after_delete = 'after_delete'
@@ -48,6 +45,11 @@ signal_map = {after_insert: _signals.signal('mongo_after_insert'),
               after_delete: _signals.signal('mongo_after_delete')}
 
 
+def resolve_class(class_path):
+    module_name, class_name = class_path.rsplit('.', 1)
+    return getattr(import_module(module_name), class_name)
+
+
 class AuthenticationError(Exception):
     pass
 
@@ -76,7 +78,7 @@ class AttrDict(dict):
     Base object that represents a MongoDB document. The object will behave both
     like a dict `x['y']` and like an object `x.y`
 
-    :param initial: you can define new instance via dictionary:
+    :param initial: you can define new instance via dictionary::
                     AttrDict({'a': 'one', 'b': 'two'}) or pass data
                     in kwargs AttrDict(a='one', b='two')
     """
@@ -139,81 +141,48 @@ class AutoincrementId(SONManipulator):
         return result["next"]
 
 
-class AutoReferenceObject(AutoReference):
+class SavedObject(SONManipulator):
     """
-    Transparently reference and de-reference already saved embedded objects.
+    Сonverts saved documents into class instance, the class name with path of
+    class module keep in document :_class: parameter e.g.::
+        {'name': 'John', 'age': 18, '_class': 'my_project.account.User'} -
+        will convert into class User
 
-    This manipulator should probably only be used when the NamespaceInjector is
-    also being used, otherwise it doesn't make too much sense - documents can
-    only be auto-referenced if they have an `_ns` field.
-
-    NOTE: this will behave poorly if you have a circular reference.
+    Embedded documents will convert, if they have :_class: parameter
 
     TODO: this only works for documents that are in the same database. To fix
     this we'll need to add a DatabaseInjector that adds `_db` and then make
     use of the optional `database` support for DBRefs.
     """
 
-    def __init__(self, mongoset):
-        self.__database = mongoset.session
+    def will_copy(self):
+        return True
+
+    def _transform_value(self, value):
+        if isinstance(value, list):
+            return map(self._transform_value, value)
+
+        if isinstance(value, dict):
+            if value.get('_class'):
+                cls = resolve_class(value['_class'])
+                return cls(self._transform_dict(value))
+            return self._transform_dict(value)
+
+        return value
+
+    def _transform_dict(self, object):
+        for (key, value) in object.items():
+            object[key] = self._transform_value(value)
+        return object
 
     def transform_outgoing(self, son, collection):
-        if collection.name in autoref_collections:
-            def transform_value(value):
-                if isinstance(value, DBRef):
-                    return transform_value(self.__database.dereference(value))
-                elif isinstance(value, list):
-                    return map(transform_value, value)
-                elif isinstance(value, dict):
-                    if value.get('_ns'):
-                        cls = autoref_collections.get(value['_ns'])
-                        if cls:
-                            return cls(transform_dict(value))
-                    return transform_dict(value)
-                return value
-
-            def transform_dict(object):
-                for (key, value) in object.items():
-                    object[key] = transform_value(value)
-                return object
-            son = transform_dict(son)
-        return son
-
-
-class SavedObject(SONManipulator):
-    """
-    Takes document from db as class instance.
-
-    This manipulator should probably only be used when the NamespaceInjector is
-    also being used, otherwise it doesn't make too much sense - nested
-    documents can only be taken from db as instace if they have an `_ns` field.
-    """
-
-    def transform_outgoing(self, son, collection):
-        if collection.name in autoref_collections:
-            def transform_value(value):
-                if isinstance(value, list):
-                    return map(transform_value, value)
-                elif isinstance(value, dict):
-                    if value.get('_ns'):
-                        cls = autoref_collections.get(value['_ns'])
-                        if cls:
-                            return cls(transform_dict(value))
-                    return transform_dict(value)
-                return value
-
-            def transform_dict(object):
-                for (key, value) in object.items():
-                    object[key] = transform_value(value)
-                return object
-            son = transform_dict(son)
-        return son
+        return self._transform_value(son)
 
 
 class MongoCursor(Cursor):
     """
-    A cursor that will return an instance of :param as_class: with
-    provided :param _lang: instead of dict
+    A cursor that will return an instance of :as_class: parameter with
+    provided :_lang: parameter instead of dict type
     """
     def __init__(self, *args, **kwargs):
         self._lang = kwargs.pop('_lang')
@@ -338,6 +307,9 @@ class ModelType(type):
     def __new__(cls, name, bases, dct):
         structure = dct.get('structure')
 
+        if structure is not None:
+            structure.allow_extra('_class', '_id', '_ns', '_int_id')
+
         # inheritance from abstract models:
         for model in bases:
 
@@ -369,7 +341,6 @@ class ModelType(type):
                                                 set(structure.ignore))
                 elif model.structure:
                     dct['structure'] = model.structure
-
                 break
 
         #  change structure for translated fields:
@@ -402,10 +373,6 @@ class ModelType(type):
         cls._protected_field_names = list(protected_field_names.union(*names))
 
         if not cls.__abstract__:
-            # add model into DBrefs register:
-            if cls.use_autorefs:
-                autoref_collections.__setitem__(cls.__collection__, cls)
-
             # add model into autoincrement_id register:
             if cls.inc_id:
                 inc_collections.add(cls.__collection__)
@@ -497,6 +464,9 @@ class Model(AttrDict):
     def __init__(self, initial=None, **kwargs):
         self.from_db = kwargs.pop('from_db', False)
         self._lang = kwargs.pop('_lang', self._fallback_lang)
+        if not self.from_db:
+            self._class = ".".join([self.__class__.__module__,
+                                    self.__class__.__name__])
         dct = kwargs.copy()
 
         if initial and isinstance(initial, dict):
@@ -713,14 +683,16 @@ class MongoSet(object):
         """
         if not hasattr(self, "db"):
             self.db = self.connection[self.app.config['MONGODB_DATABASE']]
+            # we need namespaces in any case
+            self.db.add_son_manipulator(NamespaceInjector())
+
             if self.app.config['MONGODB_AUTOREF']:
-                self.db.add_son_manipulator(NamespaceInjector())
-                self.db.add_son_manipulator(AutoReferenceObject(self))
-            else:
-                self.db.add_son_manipulator(NamespaceInjector())
-                self.db.add_son_manipulator(SavedObject())
+                self.db.add_son_manipulator(AutoReference(self.db))
+
             if self.app.config['MONGODB_AUTOINCREMENT']:
                 self.db.add_son_manipulator(AutoincrementId())
+
+            self.db.add_son_manipulator(SavedObject())
         return self.db
 
     def clear(self):
